package Test::Darcs;

use Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(
    &darcs
    &echo_to_darcs
    $DARCS
);

use strict;


=head1 NAME

Test::Darcs - functions to help testing darcs

=head1 SYNOPSIS

  use Test::More 'no_plan';
  use Test::Darcs;

  darcs 'init';

=head1 DESCRIPTION

Utility functions to help in the testing of darcs.

=head2 Functions

All functions here are exported by default.

=head3 darcs

  my $output = darcs @commands;

Runs darcs with the given @commands returning STDOUT and STDERR
combined.  Similar to:

    my $output = `darcs @commands 2>&1`;

but potentially more portable.

By default the darcs used is the one sitting in the source directory.
This can be overridden using the DARCS environment variable.

The exit code of the darcs command is available as C<$?>.

=cut

use vars qw/$DARCS/;
sub _find_darcs {
    return $DARCS if defined $DARCS;

    my $darcs = $ENV{DARCS} || "$ENV{PWD}/../darcs";
    die "darcs not found as $darcs" unless -x $darcs;
    $DARCS = $darcs;

    return $DARCS;
}

sub darcs (@) {
    my @commands = @_;
    my $darcs = _find_darcs;
    return `$darcs @commands 2>&1`;
}

=head2 echo_to_darcs()

 my $out = echo_to_darcs('y', "pull -a");

This effectively pipes the first argument to a darcs command
given as the second argument. Similiar to:

 my $out = `echo 'y' | darcs pull \a`;

but potentially more portable.

It's only good for sending one bit of input, not for a truly interactive
session.

=cut

sub echo_to_darcs {
    my($input, $command) = @_;

    my $darcs = _find_darcs;

    local(*READ, *WRITE);
    use IPC::Open2;
    open2(*READ, *WRITE, "$darcs $command");
    print WRITE "$input\n";
    close WRITE;

    my $output = join '', <READ>;

    close READ;

    return $output;
}



=head1 ENVIRONMENT

=head3 DARCS

darcs() normally uses the copy of darcs in the source directory but if
DARCS is set it will use that copy of darcs instead.

=cut

1;
